import pytest
from kibitzr.notifier.gotify import GotifyNotify

from ...compat import mock
from ...helpers import SettingsMock


@pytest.fixture()
def settings():
    """Override native settings singleton with empty one"""
    return SettingsMock.instance()


def test_gotify_sample(settings):
    settings.creds.update({
        'gotify': {
            'url': 'http://localhost:8080',
            'token': '123456',
        }
    })
    value = {}
    notify = GotifyNotify(value=value)
    with mock.patch.object(notify.session, 'post') as fake_post:
        notify('report')
    fake_post.assert_called_once_with(
        url='http://localhost:8080/message',
        json={
            'title': 'kibitzr',
            'message': 'report',
            'priority': 4
        },
        params={'token': '123456'},
        verify=True
    )


def test_gotify_no_verification(settings):
    settings.creds.update({
        'gotify': {
            'url': 'https://localhost:8080',
            'token': '123456',
            'verify': False,
        }
    })
    value = {}
    notify = GotifyNotify(value=value)
    with mock.patch.object(notify.session, 'post') as fake_post:
        notify('report')
    fake_post.assert_called_once_with(
        url='https://localhost:8080/message',
        json={
            'title': 'kibitzr',
            'message': 'report',
            'priority': 4,
        },
        params={'token': '123456'},
        verify=False
    )


def test_gotify_override_defaults(settings):
    settings.creds.update({
        'gotify': {
            'url': 'http://localhost:8080',
            'token': '123456',
        }
    })
    value = {'title': 'Important Notification', 'priority': 7}
    notify = GotifyNotify(value=value)
    with mock.patch.object(notify.session, 'post') as fake_post:
        notify('report')
    fake_post.assert_called_once_with(
        url='http://localhost:8080/message',
        json={
            'title': 'Important Notification',
            'message': 'report',
            'priority': 7
        },
        params={'token': '123456'},
        verify=True
    )



from .factory import fetcher_factory  # noqa


def persistent_firefox():
    from .browser.fetcher import persistent_firefox as real_persistent_firefox  # pylint: disable=import-outside-toplevel
    real_persistent_firefox()


def cleanup_fetchers():
    from .browser.launcher import cleanup  # pylint: disable=import-outside-toplevel
    cleanup()



import pytest
from freezegun import freeze_time

from kibitzr.timeline import Timeline, TimelineRule
from kibitzr.exceptions import ConfigurationError


def test_dummy_schedule(app, settings, check_noop):
    check_noop.side_effect = interrupt_on_nth_call(app, 2)
    settings.checks.append({
        'name': 'A',
        'script': {'python': 'ok, content = True, "ok"'},
        'schedule': [TimelineRule(interval=0, unit='seconds', at=None)],
    })
    assert app.run() == 1
    assert check_noop.call_count == 2


def interrupt_on_nth_call(app, n):
    def interrupter():
        if interrupter.n > 1:
            interrupter.n -= 1
        else:
            app.on_interrupt()
        return None
    interrupter.n = n
    return interrupter


def test_period_parse():
    rule = Timeline.parse_check({'period': '1 hour'})
    assert rule == [TimelineRule(interval=3600, unit='seconds', at=None)]


def test_empty_period():
    rule = Timeline.parse_check({'name': 'x'})
    assert rule == [TimelineRule(interval=300, unit='seconds', at=None)]


def test_sample_schedule():
    rule = Timeline.parse_check({
        'schedule': {
            'every': 1,
            'unit': 'week',
            'at': '07:40',
        }
    })
    assert rule == [TimelineRule(interval=1, unit='week', at='07:40')]


def test_invalid_schedule():
    with pytest.raises(ConfigurationError):
        Timeline.parse_check({
            'name': 'x',
            'schedule': 'daily',
        })


def test_invalid_unit():
    with pytest.raises(ConfigurationError):
        Timeline.parse_check({
            'name': 'x',
            'schedule': {
                'every': 2,
                'unit': 'olympics'
            },
        })


def test_invalid_every():
    with pytest.raises(ConfigurationError):
        print(Timeline.parse_check({
            'name': 'x',
            'schedule': {
                'every': 'single',  # every must be integer
                'unit': 'minute',
            },
        }))


def test_invalid_time():
    with pytest.raises(ConfigurationError):
        print(Timeline.parse_check({
            'name': 'x',
            'schedule': {
                'every': 'day',
                'at': 'noon',
            },
        }))


def test_schedule_daily_with_time():
    timeline = Timeline()
    timeline.schedule_checks([DummyCheck([
        TimelineRule(1, 'day', '12:34')
    ])])
    next_run = timeline.scheduler.jobs[0].next_run
    assert next_run.hour == 12
    assert next_run.minute == 34


@freeze_time("1945-05-09 12:34:56")
def test_schedule_daily_uses_current_time_by_default():
    timeline = Timeline()
    timeline.schedule_checks([DummyCheck([
        TimelineRule(1, 'day', None)
    ])])
    next_run = timeline.scheduler.jobs[0].next_run
    assert next_run.hour == 12
    assert next_run.minute == 34


class DummyCheck:
    def __init__(self, rules):
        self.conf = {
            'name': 'dummy',
            'schedule': rules,
        }

    def check(self):
        pass



"""
Kibitzr accepts shell scripts in 3 places:

* Fetch
* Transform
* Notify

Execute code from notifier with transformation result passed via stdin.

Here is a simplistic example of ``kibitzr.yml`` file, that uses all three:

.. code-block:: yaml

    checks:
      - name: Shell example
        script: |
          for i in seq 1 3
          do
              echo "Number $i"
          done
        transform:
          - shell: grep 2
        notify:
          - shell: tac

Let's break it down.

.. _shell-fetcher:

Shell Fetcher
-------------

If ``script``'s value is a string, it will be used as shell script.
Alternatively ``script`` can hold a dictionary of one item.
Item's key can be ``shell`` (or ``python`` for :ref:`Python fetcher <python>`).
If ``script``'s only key is ``shell``, then it's value will be
executed as a Shell script.
Under Linux, executor is ``bash``, under Windows - ``cmd.exe``.
Script is an arbitrary shell code. It's output will be passed to transforms.
If exit code is not zero, check will be aborted.
Shell scripts don't have access to credentials, but inherit Kibitzr environment.

Shell Transform and Notifier
----------------------------

Transform and notifier are similar to fetcher. Except that they receive content via stdin,
and notifier's stdout is ignored.

Example
-------

Returning to the example, execution will go as follows:

.. code-block:: bash

    [DEBUG] kibitzr.conf: Loading settings from /home/deminp/kibitzr/tmp/kibitzr.yml
    [INFO] kibitzr.fetcher.loader: Fetching 'Shell' using script
    [DEBUG] kibitzr.bash: Saving code to '/tmp/tmpTTPSxA.bat'
    [DEBUG] kibitzr.bash: Launching script '/tmp/tmpTTPSxA.bat'
    [DEBUG] kibitzr.bash: Command exit code: 0
    [DEBUG] kibitzr.bash: Command stdout: Number 1
    Number 2
    Number 3

    [DEBUG] kibitzr.bash: Command stderr:
    [DEBUG] kibitzr.bash: Saving code to '/tmp/tmpV4Grg8.bat'
    [DEBUG] kibitzr.bash: Launching script '/tmp/tmpV4Grg8.bat'
    [DEBUG] kibitzr.bash: Command exit code: 0
    [DEBUG] kibitzr.bash: Command stdout: Number 2
    [DEBUG] kibitzr.bash: Command stderr:
    [DEBUG] kibitzr.notifier.factory: Sending report: u'Number 2'
    [DEBUG] kibitzr.bash: Saving code to '/tmp/tmpm6sRVx.bat'
    [DEBUG] kibitzr.bash: Launching script '/tmp/tmpm6sRVx.bat'
    [DEBUG] kibitzr.bash: Command exit code: 0
    [DEBUG] kibitzr.bash: Command stdout: 2 rebmuN
    [DEBUG] kibitzr.bash: Command stderr:

Fetcher script produced output:

.. code-block:: bash

    Number 1
    Number 2
    Number 3

Shell transform filtered lines that contain 2:

.. code-block:: bash

    Number 2

Notifier echoed reversed line:

.. code-block:: bash

    2 rebmuN

Notifier's stdout is ignored, so we don't see it along Kibitzr output.

And here is what happens when shell script produces error:

.. code-block:: bash

    $ cat kibitzr.yml
    checks:
      - name: Shell
        script: ls /non-existing
        notify:
          - shell: rev

    $ kibitzr -l debug once
    [DEBUG] kibitzr.conf: Loading settings from /home/deminp/kibitzr/tmp/kibitzr.yml
    [INFO] kibitzr.fetcher.loader: Fetching 'Shell' using script
    [DEBUG] kibitzr.bash: Saving code to '/tmp/tmpyNakOP.bat'
    [DEBUG] kibitzr.bash: Launching script '/tmp/tmpyNakOP.bat'
    [ERROR] kibitzr.bash: Command exit code: 2
    [ERROR] kibitzr.bash: Command stdout:
    [ERROR] kibitzr.bash: Command stderr: ls: cannot access '/non-existing': No such file or directory
    [DEBUG] kibitzr.transformer.factory: Notifying on error
    [DEBUG] kibitzr.notifier.factory: Sending report: u"ls: cannot access '/non-existing': No such file or directory"
    [DEBUG] kibitzr.bash: Saving code to '/tmp/tmpqdZwKI.bat'
    [DEBUG] kibitzr.bash: Launching script '/tmp/tmpqdZwKI.bat'
    [DEBUG] kibitzr.bash: Command exit code: 0
    [DEBUG] kibitzr.bash: Command stdout: yrotcerid ro elif hcus oN :'gnitsixe-non/' ssecca tonnac :sl
    [DEBUG] kibitzr.bash: Command stderr:
"""
import os
import logging
import tempfile
import contextlib

import six


logger = logging.getLogger(__name__)


def execute_bash(code, stdin=None):
    if os.name == 'nt':
        executor = WindowsExecutor
    else:
        executor = BashExecutor
    return executor(code).execute(stdin)


class BashExecutor:

    EXECUTABLE = "bash"
    ARGS = []

    def __init__(self, code):
        self.code = code

    def execute(self, stdin=None):
        if stdin is not None and stdin.strip():
            stdin = stdin.encode("utf-8")
            with self.temp_file() as filename:
                ok, result = self.run_scipt(filename, stdin)
            return self.make_report(ok, result)
        else:
            logger.info("Skipping execution with empty content")
            return True, stdin

    @contextlib.contextmanager
    def temp_file(self):
        """
        Create temporary file with code and yield its path.
        Works both on Windows and Linux
        """
        with tempfile.NamedTemporaryFile(suffix='.bat', delete=False) as fp:
            try:
                logger.debug("Saving code to %r", fp.name)
                fp.write(self.code.encode('utf-8'))
                fp.close()
                yield fp.name
            finally:
                os.remove(fp.name)

    @classmethod
    def run_scipt(cls, name, stdin):
        from kibitzr.compat import sh  # pylint: disable=import-outside-toplevel
        logger.debug("Launching script %r", name)
        try:
            args = cls.ARGS + [name]
            return True, sh.Command(cls.EXECUTABLE)(*args, _in=stdin)
        except sh.ErrorReturnCode as exc:
            return False, exc

    @staticmethod
    def make_report(ok, result):
        stdout = ensure_text(result.stdout)
        stderr = ensure_text(result.stderr)
        if ok:
            log = logger.debug
            report = stdout
        else:
            log = logger.error
            report = stderr
        if hasattr(result, 'exit_code'):
            log("Command exit code: %r", result.exit_code)
        log("Command stdout: %s", stdout)
        log("Command stderr: %s", stderr)
        return ok, report


class WindowsExecutor(BashExecutor):

    EXECUTABLE = "cmd.exe"
    ARGS = ["/Q", "/C"]

    @classmethod
    def run_scipt(cls, name, stdin):
        return BashExecutor.run_scipt(
            name,
            stdin.decode("utf-8"),
        )


def ensure_text(text):
    if not isinstance(text, six.text_type):
        return text.decode('utf-8')
    else:
        return text



import pytest
from kibitzr.app import Application
from ..helpers import SettingsMock


@pytest.fixture(scope="function")
def settings(request):
    """Override native settings singleton with empty one"""
    instance = SettingsMock.instance()
    request.addfinalizer(SettingsMock.dispose)
    return instance


@pytest.fixture(autouse=True)
def check_noop(mocker):
    """Cancel all side-effects from checks"""
    return mocker.patch("kibitzr.app.Checker.check", spec=True)


@pytest.fixture
def app():
    return Application()



# -*- coding: utf-8 -*-
from bs4 import BeautifulSoup

from .helpers import run_transform, HTML


def test_tag_selector():
    ok, content = run_transform('tag', 'a', HTML)
    assert ok is True
    assert content == '<a href="page.html" id="page-link">Page</a>'


def test_css_selector():
    ok, content = run_transform('css', 'body h2.nav a#page-link', HTML)
    assert ok is True
    assert content == '<a href="page.html" id="page-link">Page</a>'


def test_css_selector_all():
    ok, content = run_transform('css-all', 'div', HTML)
    assert ok is True
    assert prettify(content) == u'\n'.join([
        u'<div id="content">',
        u' Привет, Мир!',
        u'</div>',
        u'',
        u'<div class="footer">',
        u' Footer content',
        u'</div>',
    ])


def test_extract_test():
    ok, content = run_transform('text', None, HTML)
    assert ok is True
    assert content.strip() == u"\n".join([
        u"Page",
        u"Привет, Мир!",
        u"Footer content",
    ])


def prettify(html):
    return "\n".join(
        child.prettify()
        for child in BeautifulSoup(html, "lxml").html.body.children
    )



import pytest
from twilio.rest import Client

from kibitzr.notifier.twilio import notify

from ...compat import mock
from ...helpers import SettingsMock


@pytest.fixture(name='settings')
def settings_fixture():
    """Override native settings singleton with empty one"""
    return SettingsMock.instance()


def test_twilio_sample(settings):
    settings.creds.update(
        {
            "twilio": {
                "account_sid": "sid",
                "auth_token": "token",
                "phone_number": "phone_number",
            }
        }
    )
    fake_client = mock.Mock(auto_spec=Client)
    notify(
        "report",
        notifier_conf={"recipients": ["recipient"]},
        client_factory=fake_client,
    )
    fake_client.assert_called_once_with("sid", "token")
    client_instance = fake_client.return_value
    client_instance.messages.create.assert_called_once_with(
        body="report",
        from_="phone_number",
        to="recipient",
    )



from kibitzr.notifier import notify_factory

from ...helpers import stash_mock


def make_stash(**kwargs):
    return notify_factory({'notify': [{'stash': dict(**kwargs)}]})


def test_stash_is_persistent():
    with stash_mock() as stash:
        notifier = make_stash(key='{{content}}')
        notifier.notify('test')
        assert stash.read() == {'key': 'test'}



"""
Kibitzr supports a number of fetchers and the list can be extended by plugins.

In the nutshell each fetcher takes conf and returns a tuple: ok, content.
where ok is boolean meaning success, and content is a Unicode string with fetch result.

Each fetcher has a promoter.
Promoter knows if the fetcher is applicable for given conf and
sets a priority for conflict resolution.

Promoter instance is initialized with conf and delegates calls to fetcher.
"""

from kibitzr.exceptions import ConfigurationError

from .loader import load_promoters


_PROMOTERS = None


def fetcher_factory(conf):
    """Return initialized fetcher capable of processing given conf."""
    global _PROMOTERS  # pylint: disable=global-statement
    applicable = []
    if not _PROMOTERS:
        _PROMOTERS = load_promoters()
    for promoter in _PROMOTERS:
        if promoter.is_applicable(conf):
            applicable.append((promoter.PRIORITY, promoter))
    if applicable:
        best_match = sorted(applicable, reverse=True)[0][1]
        return best_match(conf)
    raise ConfigurationError(
        'No fetcher is applicable for "{conf["name"]}"'
    )



import contextlib

from kibitzr.conf import (
    ReloadableSettings,
    PlainYamlCreds,
)
from ..compat import mock


@contextlib.contextmanager
def patch_creds(data):
    with patch_source(PlainYamlCreds, 'open_creds', data) as fake_method:
        yield fake_method


@contextlib.contextmanager
def patch_conf(data):
    with patch_source(ReloadableSettings, 'open_conf', data) as fake_method:
        yield fake_method


@contextlib.contextmanager
def patch_source(klass, method, data):
    fake_file = mock.mock_open(read_data=data)
    with mock.patch.object(klass,
                           method,
                           fake_file,
                           create=True) as fake_method:
        yield fake_method



import pytest
from kibitzr.conf import (
    settings,
    ReloadableSettings,
    ConfigurationError,
)
from kibitzr.timeline import TimelineRule
from ..compat import mock
from .helpers import patch_conf, patch_creds


sample_conf = """
notifiers:
    slack:
        url: XXX

templates:
    teamcity:
        transforms:
            - text
            - changes: verbose
        scenario:
            login
        notify:
            - smtp: kibitzrrr@gmail.com
        period: 3600

scenarios:
    login: |
        driver.find_element_by_css_selector(".login").click()

checks:
    - name: Project Build Status
      url: http://teamcity/build/id
      template: teamcity
      period: 60

    - batch: WordPress {0} Plugin
      url-pattern: "http://wordpress/{0}"
      notify:
        - mailgun
      items:
        - A
        - B

    - name: Alarm clock
      url: https://www.worldtimeserver.com/current_time_in_US-NY.aspx
      transform:
          - css: "span#theTime"
          - text
      notify:
          - python: print(content)
      schedule:
            every: day
            at: "6:30"

    - name: Noon alarm
      url: https://www.worldtimeserver.com/current_time_in_US-NY.aspx
      transform:
          - css: "span#theTime"
          - text
      notify:
          - python: print(content)
      schedule:
            every: 1
            unit: day
            at: "20:30"

    - name: Crazy scheduling alarm
      url: https://www.worldtimeserver.com/current_time_in_US-NY.aspx
      transform:
          - css: "span#theTime"
          - text
      notify:
          - python: print(content)
      schedule:
          - every: 1
            unit: days
            at: "15:30"
          - every: hour
          - every: 4
            unit: minutes
          - every: saturday
            at: "11:17"
"""

sample_creds = """
mailgun:
    url: ZZZ
"""


checks = [
    {
        'name': 'Project Build Status',
        'url': 'http://teamcity/build/id',
        'transforms': ['text', {'changes': 'verbose'}],
        'scenario': 'driver.find_element_by_css_selector(".login").click()\n',
        'notify': [{'smtp': 'kibitzrrr@gmail.com'}],
        'schedule': [TimelineRule(interval=60, unit='seconds', at=None)]
    },
    {
        'name': 'WordPress A Plugin',
        'url': "http://wordpress/A",
        'notify': ['mailgun'],
        'schedule': [TimelineRule(interval=300, unit='seconds', at=None)]
    },
    {
        'name': 'WordPress B Plugin',
        'url': "http://wordpress/B",
        'notify': ['mailgun'],
        'schedule': [TimelineRule(interval=300, unit='seconds', at=None)]
    },
    {
        'name': 'Alarm clock',
        'url': "https://www.worldtimeserver.com/current_time_in_US-NY.aspx",
        'transform': [{'css': 'span#theTime'}, 'text'],
        'notify': [{'python': 'print(content)'}],
        'schedule': [TimelineRule(interval=1, unit='day', at='6:30')]
    },
    {
        'name': 'Noon alarm',
        'url': "https://www.worldtimeserver.com/current_time_in_US-NY.aspx",
        'transform': [{'css': 'span#theTime'}, 'text'],
        'notify': [{'python': 'print(content)'}],
        'schedule': [TimelineRule(interval=1, unit='day', at='20:30')]
    },
    {
        'name': 'Crazy scheduling alarm',
        'url': "https://www.worldtimeserver.com/current_time_in_US-NY.aspx",
        'transform': [{'css': 'span#theTime'}, 'text'],
        'notify': [{'python': 'print(content)'}],
        'schedule': [
            TimelineRule(interval=1, unit='days', at='15:30'),
            TimelineRule(interval=1, unit='hour', at=None),
            TimelineRule(interval=4, unit='minutes', at=None),
            TimelineRule(interval=1, unit='saturday', at='11:17')
        ]
    },
]

creds = {
    'mailgun': {'url': 'ZZZ'},
}

notifiers = {
    'slack': {'url': 'XXX'},
}


def test_complex_conf_sample():
    with patch_conf(sample_conf):
        with patch_creds(sample_creds):
            conf = ReloadableSettings('::')
    assert conf.checks == checks
    assert conf.creds['mailgun']['url'] == 'ZZZ'


@mock.patch("kibitzr.conf.os.path.exists", return_value=False)
def test_missing_config_raises_configuration_error(exists):
    with pytest.raises(ConfigurationError):
        settings()


def test_reread():
    conf1 = (
        "checks:\n"
        "  - name: Name\n"
        "    url: URL\n"
    )
    conf2 = conf1 + "    period: 60\n"
    with patch_creds(""):
        with patch_conf(conf1):
            conf = ReloadableSettings('::')
            # Config didn't change:
            assert not conf.reread()
        with patch_conf(conf2):
            # Config changed:
            assert conf.reread()
    assert conf.checks == [{
        'name': 'Name',
        'url': 'URL',
        'schedule': [TimelineRule(interval=60, unit='seconds', at=None)]
    }]


def test_name_from_url_population():
    conf = "checks: [{url: 'http://example.com/page_name'}]"
    with patch_conf(conf):
        conf = ReloadableSettings('::')
    assert conf.checks[0]['name'] == 'http-example-com-page_name'


def test_unnamed_check():
    conf = "checks: [{period: 1}]"
    with patch_conf(conf):
        conf = ReloadableSettings('::')
    assert conf.checks == [{
        'name': 'Unnamed check 1',
        'schedule': [TimelineRule(interval=1, unit='seconds', at=None)]
    }]



import json

from .webhook import WebHookNotify, webhook_factory


class DiscordNotify(WebHookNotify):
    CREDS_KEY = 'discord'

    def configure_session(self):
        self.session.headers.update({'Content-Type': 'application/json'})

    def payload(self, report):
        return json.dumps({"content": report})


notify_factory = webhook_factory(DiscordNotify)



import re
import os
import logging
import json
import functools

import six

from kibitzr.stash import LazyStash
from .html import deep_recursion, SoupOps
from .xpath import parse_html, serialize_xpath_results

logger = logging.getLogger(__name__)


class JinjaTransform:

    def __init__(self, code, conf):
        from jinja2 import Environment  # pylint: disable=import-outside-toplevel
        environment = Environment()
        environment.filters['text'] = text_filter
        environment.filters['int'] = int_filter
        environment.filters['float'] = float_filter
        environment.filters['dollars'] = dollars_filter
        self.template = environment.from_string(code)
        self.conf = conf

    def render(self, content, context=None):
        from jinja2 import TemplateError  # pylint: disable=import-outside-toplevel
        try:
            return True, self.template.render(context or self.context(content))
        except TemplateError:
            logger.warning("Jinja render failed", exc_info=True)
            return False, None
    __call__ = render

    def context(self, content):
        html = LazyHTML(content)
        xml = LazyXML(content)
        return {
            'conf': self.conf,
            'stash': LazyStash(),
            'content': content,
            'lines': content.splitlines(),
            'json': LazyJSON(content),
            'css': html.css,
            'xpath': xml.xpath,
            'env': os.environ,
        }


RE_NOT_FLOAT = re.compile(r'[^0-9\.]')


def ignore_cast_error(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except (TypeError, ValueError):
            logger.warning("Invalid value passed in Jinja transform",
                           exc_info=True)
            return None
    return wrapper


@ignore_cast_error
def dollars_filter(number):
    sign = '-' if number < 0 else ''
    return f"{sign}${abs(number):,}"


@ignore_cast_error
def int_filter(text):
    return int(text)


@ignore_cast_error
def float_filter(text):
    return float(RE_NOT_FLOAT.sub('', text))


def text_filter(html):
    if isinstance(html, list):
        html = "".join(html)
    ok, content = SoupOps.extract_text(html)
    if ok:
        return content
    else:
        raise RuntimeError("Extract text failed")


class LazyJSON:
    def __init__(self, content):
        self.text = content
        self._json = None

    @property
    def json(self):
        if self._json is None:
            self._json = json.loads(self.text)
        return self._json

    def __getitem__(self, key):
        return self.json[key]


class LazyHTML:
    def __init__(self, content):
        self.html = content
        self._soup = None

    @property
    def soup(self):
        from bs4 import BeautifulSoup  # pylint: disable=import-outside-toplevel
        if self._soup is None:
            self._soup = BeautifulSoup(self.html, "html.parser")
        return self._soup

    def css(self, selector):
        with deep_recursion():
            elements = self.soup.select(selector)
            result = [six.text_type(x)
                      for x in elements]
            return result


class LazyXML:
    def __init__(self, content):
        self.xml = content
        self._root = None

    @property
    def root(self):
        if self._root is None:
            self._root = parse_html(self.xml)
        return self._root

    def xpath(self, selector):
        elements = self.root.xpath(selector)
        return serialize_xpath_results(elements, select_all=True)


def register():
    return {
        'jinja': JinjaTransform,
    }



from ..compat import mock


def test_loop_aborts_without_checks(app, settings):
    assert app.run() == 1


def test_main_executes_all_checks_before_loop(app, settings):
    with mock.patch.object(app, "check_forever", side_effect=app.on_interrupt) as the_loop:
        settings.checks.append({
            'name': 'A',
            'script': {'python': 'ok, content = True, "ok"'}
        })
        assert app.run() == 1
    assert the_loop.call_count == 1
    assert the_loop.call_args[0][0][0].check.call_count == 1


def test_main_filters_names(app, settings):
    with mock.patch.object(app, "check_forever", side_effect=app.on_interrupt) as the_loop:
        settings.checks.extend([
            {'name': 'A', 'url': 'A'},
            {'name': 'B', 'url': 'B'},
        ])
        assert app.run(names=['B']) == 1
    assert the_loop.call_count == 1
    assert the_loop.call_args[0][0][0].check.call_count == 1



# -*- coding: utf-8 -*-
from kibitzr.transformer import transform_factory


HTML = u"""<?xml version="1.0" encoding="utf-8"?>
<html>
    <body>
        <h2 class="header nav">
            <a href="page.html" id="page-link">Page</a>
        </h2>
        <div id="content">
            Привет, Мир!
        </div>
        <div class="footer">
            Footer content
        </div>
    </body>
</html>
"""


def run_transform(key, value, content):
    pipeline = transform_factory({'transform': [{key: value}]})
    return pipeline(True, content)



import logging

import six

from .utils import bake_parametrized


logger = logging.getLogger(__name__)


def parse_html(html):
    """
    Returns `html` parsed with lxml.

    :param html: Unicode content
    """
    # pylint: disable=import-outside-toplevel
    from defusedxml import lxml as dlxml
    from lxml import etree

    # lxml requires argument to be bytes
    # see https://github.com/kibitzr/kibitzr/issues/47
    encoded = html.encode('utf-8')
    return dlxml.fromstring(encoded, parser=etree.HTMLParser())


def serialize_xpath_results(xpath_results, select_all):
    """
    Serializes results of xpath evaluation.

    :param xpath_results: Results of xpath evaluation.
    See: https://lxml.de/xpathxslt.html#xpath-return-values

    :param select_all: True to get all matches
    """
    # pylint: disable=import-outside-toplevel
    import re
    from defusedxml import lxml as dlxml

    if isinstance(xpath_results, list):
        if select_all is False:
            xpath_results = xpath_results[0:1]
    else:
        xpath_results = [xpath_results]

    results = []
    for r in xpath_results:
        # namespace declarations
        if isinstance(r, tuple):
            results.append(f'{r[0]}="{r[1]}"')
        # an element
        elif hasattr(r, 'tag'):
            results.append(
                re.sub(r'\s+', ' ',
                       dlxml.tostring(r, method='html', encoding='unicode'))
            )
        else:
            results.append(r)

    return u"\n".join(six.text_type(x).strip() for x in results)


def xpath_selector(selector, html, select_all):
    """
    Returns Xpath match for `selector` within `html`.

    :param selector: XPath string
    :param html: Unicode content
    :param select_all: True to get all matches
    """
    root = parse_html(html)
    xpath_results = root.xpath(selector)

    if not xpath_results:
        logger.warning('XPath selector not found: %r', selector)
        return False, html
    return True, serialize_xpath_results(xpath_results, select_all)


def register():
    """
    Return dictionary of transform factories
    """
    registry = {
        'xpath': bake_parametrized(xpath_selector, select_all=False),
        'xpath-all': bake_parametrized(xpath_selector, select_all=True)
    }
    return registry



try:
    import sh  # noqa
except ImportError:
    # Windows alternative:
    import pbs as sh  # noqa

try:
    from smtplib import SMTPNotSupportedError  # noqa
except ImportError:
    # SMTPNotSupportedError is new in 3.5
    from smtplib import SMTPException as SMTPNotSupportedError  # noqa


__all__ = (
    'sh',
    'SMTPNotSupportedError',
)



import logging
import urllib3

from ..conf import settings


logger = logging.getLogger(__name__)


class GotifyNotify:
    def __init__(self, value):
        import requests  # pylint: disable=import-outside-toplevel
        self.session = requests.Session()
        gotify_creds = settings().creds['gotify']
        self.url = gotify_creds['url'].rstrip('/') + '/message'
        self.token = gotify_creds['token']
        self.verify = gotify_creds.get('verify', True)
        if not self.verify:
            # Since certificate validation was disabled, do not show the warning
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        if value is None:
            value = {}
        self.title = value.get('title', 'kibitzr')
        self.priority = value.get('priority', 4)

    def post(self, report):
        logger.info("Executing Gotify notifier")
        response = self.session.post(
            url=self.url,
            json=self.json(report),
            params=self.params(),
            verify=self.verify
        )
        logger.debug(response.text)
        response.raise_for_status()
        return response
    __call__ = post

    def params(self):
        return {
            'token': self.token
        }

    def json(self, report):
        return {
            'title': self.title,
            'message': report,
            'priority': self.priority
        }


def notify_factory(conf, value):
    return GotifyNotify(value)



import functools
import logging

import six

from ..conf import settings

logger = logging.getLogger(__name__)


def notify_factory(conf, value):
    @functools.wraps(notify)
    def baked_notify(report):
        from twilio.rest import Client  # pylint: disable=import-outside-toplevel
        return notify(
            report=report,
            notifier_conf=value,
            client_factory=Client,
        )

    del conf
    return baked_notify


def notify(report, notifier_conf, client_factory):
    logger.info("Executing Twilio notifier")
    credentials = settings().creds.get("twilio", {})
    account_sid = credentials.get("account_sid", "")
    auth_token = credentials.get("auth_token", "")
    src_phone_number = credentials.get("phone_number", "")
    client = client_factory(account_sid, auth_token)
    try:
        recipients = notifier_conf["recipients"]
    except (TypeError, KeyError):
        recipients = notifier_conf
    if isinstance(recipients, six.string_types):
        recipients = [recipients]
    if not src_phone_number:
        src_phone_number = recipients[0]
    for recipient in recipients:
        client.messages.create(
            body=report,
            from_=src_phone_number,
            to=recipient,
        )



import logging
import functools
from smtplib import SMTP

import six

from ..conf import settings
from ..compat import SMTPNotSupportedError


logger = logging.getLogger(__name__)


def notify_factory(conf, value):
    @functools.wraps(notify)
    def baked_notify(report):
        return notify(
            conf=conf,
            report=report,
            notifier_conf=value,
        )
    return baked_notify


def notify(conf, report, notifier_conf):
    logger.info("Executing SMTP notifier")
    credentials = settings().creds.get('smtp', {})
    user = credentials.get('user')
    password = credentials.get('password', '')
    host = credentials.get('host', 'localhost')
    port = credentials.get('port', 25)
    try:
        recipients = notifier_conf['recipients']
    except (TypeError, KeyError):
        recipients = notifier_conf
    if isinstance(recipients, six.string_types):
        recipients = [recipients]
    try:
        subject = notifier_conf['subject']
    except (TypeError, KeyError):
        subject = "Kibitzr update for " + conf['name']
    if not user:
        user = recipients[0]
    send_email(
        user=user,
        password=password,
        recipients=recipients,
        subject=subject,
        body=report,
        host=host,
        port=port,
    )


def send_email(user, password, recipients, subject, body, host, port):
    # pylint: disable=too-many-arguments
    message = (
        f"From: {user}\r\n"
        f"To: {', '.join(recipients)}\r\n"
        f"Subject: {subject}\r\n\r\n"
        f"{body}\r\n"
    )
    try:
        server = SMTP(host, port)  # ("smtp.gmail.com", 587)
        server.ehlo()
        try:
            server.starttls()
            if user and password:
                server.login(user, password)
        except SMTPNotSupportedError:
            # Localhost SMTP servers don't use authentication
            pass
        server.sendmail(user, recipients, message.encode("utf-8"))
        server.close()
        logger.debug('Successfully sent the mail')
    except:
        logger.exception("Failed to send the e-mail")



from __future__ import absolute_import
import logging
import asyncio

from ..conf import settings


logger = logging.getLogger(__name__)


class TelegramBot:
    def __init__(self, chat_id=None, split_on=None):
        from telegram import Bot  # pylint: disable=import-outside-toplevel
        telegram_creds = settings().creds['telegram']
        token = telegram_creds['token']
        if chat_id is not None:
            self._chat_id = chat_id
        else:
            self._chat_id = telegram_creds.get('chat')
        self.split_on = split_on
        self.bot = Bot(token=token)

    @property
    def chat_id(self):
        if self._chat_id is None:
            chat = self.bot.getUpdates(limit=1)[0].message.chat
            logger.debug("Imprinted chat id %d of type %s",
                         chat.id, chat.type)
            self._chat_id = chat.id
        return self._chat_id

    def post(self, report, **kwargs):
        del kwargs
        if self.split_on:
            report = report.split(self.split_on)
        else:
            report = [report]

        for r in report:
            # Telegram max message length is 4096 chars
            messages = [r[i:i + 4096] for i in range(0, len(r), 4096)]
            for m in messages:
                self.send_message(m)

    def send_message(self, message):
        message = self.bot.send_message(
            self.chat_id,
            message,
            parse_mode='Markdown',
        )
        asyncio.run(message)
        return message

    __call__ = post


def notify_factory(conf, value):
    del conf
    try:
        chat_id = value['chat']
    except (TypeError, KeyError):
        chat_id = value
    try:
        split_on = value['split-on']
    except (TypeError, KeyError):
        split_on = None

    return TelegramBot(chat_id=chat_id, split_on=split_on).post


def get_chat_id():
    bot = TelegramBot()
    print(bot.chat_id)



import logging


logger = logging.getLogger(__name__)


class BasePromoter:
    """
    Promoters are filtered by the result of is_applicable call,
    then sorted by PRIORITY.
    The one with highest PRIORITY is chosen for check.
    """

    PRIORITY = 10

    def __init__(self, conf):
        self.conf = conf

    @staticmethod
    def is_applicable(conf):
        """Return whether this promoter is applicable for given conf"""
        return bool(conf)

    def log_announcement(self):
        logger.info(u"Fetching %s using %s",
                    self.conf['name'],
                    self.__class__.__name__)

    def fetch(self):
        self.log_announcement()

    def __call__(self):
        return self.fetch()



# -*- coding: utf-8 -*-
import pytest

from .helpers import run_transform, HTML


@pytest.mark.parametrize("selector,expected", [
    pytest.param('boolean(//body/h2/a)', 'True', id='boolean'),
    pytest.param("count(//div)", "2.0", id="float"),
    pytest.param("string(//a)", "Page", id="string"),
    pytest.param("//body/h2/a/@href", "page.html", id="attribute"),
    pytest.param(
        '//body/h2/a[@id="page-link"]',
        '<a href="page.html" id="page-link">Page</a>',
        id='element'
    ),
    pytest.param(
        '/html/namespace::*[name()]',
        'xml="http://www.w3.org/XML/1998/namespace"',
        id='namespace'
    ),
])
def test_xpath_selector(selector, expected):
    ok, content = run_transform('xpath', selector, HTML)
    assert ok is True
    assert content == expected


def test_xpath_selector_all():
    ok, content = run_transform('xpath-all', '//div', HTML)
    assert ok is True
    assert content == u'\n'.join([
        u'<div id="content"> Привет, Мир! </div>',
        u'<div class="footer"> Footer content </div>',
    ])



import os
import logging
import signal
import time
import code
import psutil

import entrypoints

from .conf import settings, SettingsParser
from .fetcher import cleanup_fetchers, persistent_firefox
from .checker import Checker
from .bootstrap import create_boilerplate
from . import timeline


logger = logging.getLogger(__name__)


__all__ = [
    'Application',
]


class Application:

    def __init__(self):
        self.signals = {
            'reload_conf_pending': False,
            'interrupted': False,
            'open_backdoor': False,
            'orig': {
                signal.SIGINT: None,
                signal.SIGTERM: None,
            }
        }
        try:
            self.signals['orig'].update({
                signal.SIGUSR1: None,
                signal.SIGUSR2: None,
            })
        except AttributeError:
            # Unavailable on Windows
            pass

    @staticmethod
    def bootstrap():
        create_boilerplate()

    def run(self, once=False, log_level=logging.INFO, names=None):
        # Reset global state for testability:
        self.signals.update({
            'reload_conf_pending': False,
            'interrupted': False,
            'open_backdoor': False,
        })
        self.setup_logger(log_level)
        self.connect_signals()
        try:
            while True:
                if self.signals['interrupted']:
                    return 1
                if self.signals['reload_conf_pending']:
                    settings().reread()
                    self.signals['reload_conf_pending'] = False
                checkers = Checker.create_from_settings(
                    checks=settings().checks,
                    names=names
                )
                if checkers:
                    self.before_start(checkers)
                    self.execute_all(checkers)
                    if once:
                        return 0
                    self.check_forever(checkers)
                else:
                    logger.warning("No checks defined. Exiting")
                    return 1
        finally:
            cleanup_fetchers()
        return 0

    def disconnect_signals(self):
        signal.signal(signal.SIGINT, self.signals['orig'][signal.SIGINT])
        signal.signal(signal.SIGTERM, self.signals['orig'][signal.SIGTERM])
        try:
            signal.signal(signal.SIGUSR1, self.signals['orig'][signal.SIGUSR1])
            signal.signal(signal.SIGUSR2, self.signals['orig'][signal.SIGUSR2])
        except AttributeError:
            # Unavailable on Windows
            pass

    def connect_signals(self):
        self.signals['orig'][signal.SIGINT] = signal.signal(signal.SIGINT, self.on_interrupt)
        self.signals['orig'][signal.SIGTERM] = signal.signal(signal.SIGTERM, self.on_interrupt)
        try:
            self.signals['orig'][signal.SIGUSR1] = signal.signal(signal.SIGUSR1,
                                                                 self.on_reload_config)
            self.signals['orig'][signal.SIGUSR2] = signal.signal(signal.SIGUSR2, self.on_backdoor)
        except AttributeError:
            # Unavailable on Windows
            pass

    @staticmethod
    def execute_conf(conf):
        logging.basicConfig(level=logging.WARNING)
        logging.getLogger('').handlers[0].level = logging.WARNING
        checks = SettingsParser().parse_checks(conf)
        for check in checks:
            Checker(check).check()

    def run_firefox(self):
        self.setup_logger(logging.INFO)
        persistent_firefox()

    @staticmethod
    def telegram_chat():
        from .notifier.telegram import get_chat_id  # pylint: disable=import-outside-toplevel
        get_chat_id()

    @staticmethod
    def setup_logger(log_level=logging.INFO):
        logging.getLogger("").setLevel(log_level)

    def check_forever(self, checkers):
        timeline.schedule_checks(checkers)
        logger.info("Starting infinite loop")
        while not self.signals['reload_conf_pending']:
            if self.signals['interrupted']:
                break
            if self.signals['open_backdoor']:
                self.signals['open_backdoor'] = False
                code.interact(
                    banner="Kibitzr debug shell",
                    local=locals(),
                )
            timeline.run_pending()
            if self.signals['interrupted']:
                break
            time.sleep(1)

    def execute_all(self, checkers):
        for checker in checkers:
            if not self.signals['interrupted']:
                checker.check()
            else:
                break

    def on_reload_config(self, *args, **kwargs):
        del args, kwargs
        logger.info("Received SIGUSR1. Flagging configuration reload")
        self.signals['reload_conf_pending'] = True

    def on_backdoor(self, *args, **kwargs):
        del args, kwargs
        logger.info("Received SIGUSR2. Flagging backdoor to open")
        self.signals['open_backdoor'] = True

    def on_interrupt(self, *args, **kwargs):
        del args, kwargs
        if not self.signals['interrupted']:
            self.signals['interrupted'] = True
        else:
            # Third Ctrl+C to hard stop:
            self.disconnect_signals()

    def before_start(self, checkers):
        """
        Loads entry points named kibitzr.before_start
        and call each one with two arguments:

            1. Application instance;
            2. List of configured checkers
        """
        for point in entrypoints.get_group_all("kibitzr.before_start"):
            entry = point.load()
            entry(self, checkers)

    @staticmethod
    def send_reload():
        """
        Sends SIGUSR1 to all processes that execute kibitzr
        """
        user_id = os.geteuid()
        for proc in psutil.process_iter(['uids', 'name', 'pid']):
            if proc.info['name'] == "kibitzr" and proc.info['uids'][0] == user_id:
                proc.send_signal(signal.SIGUSR1)
                logger.info("Send singal SIGUSR1 to process: %r", proc.info['pid'])



#!/usr/bin/env python

import os
from setuptools import setup

with open('README.rst', encoding='utf-8') as readme_file:
    readme = readme_file.read()

try:
    # Might be missing if no pandoc installed
    with open('CHANGELOG.rst', encoding='utf-8') as history_file:
        history = history_file.read()
except IOError:
    history = ""


def read_requirements(ext: str = 'in'):
    with open(os.path.join('requirements', f'base.{ext}'), encoding='utf-8') as fp:
        lines = [line.split('#', 1)[0].strip()
                 for line in fp]
    # drop empty lines:
    return [line
            for line in lines
            if line and not line.startswith('#')]


def get_requirements(locked: bool):
    requirements = read_requirements('txt' if locked else 'in')
    if os.name == 'nt':
        # sh predecessor working under Windows:
        requirements.append('pbs')
    else:
        requirements.extend(['sh<2'])
    return requirements


setup(
    name='kibitzr',
    version='7.0.5',
    description="Self hosted web page changes monitoring",
    long_description=readme + '\n\n' + history,
    author="Peter Demin",
    author_email='kibitzrrr@gmail.com',
    url='https://github.com/kibitzr/kibitzr',
    packages=[
        'kibitzr',
    ],
    package_dir={
        'kibitzr': 'kibitzr',
    },
    entry_points={
        'console_scripts': [
            'kibitzr=kibitzr.cli:extended_cli'
        ]
    },
    include_package_data=True,
    license="MIT license",
    zip_safe=False,
    keywords='kibitzr',
    classifiers=[
        'Development Status :: 6 - Mature',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    python_requires='>=3.8',
    test_suite='tests',
    install_requires=get_requirements(locked=False),
    extras_require={
        'locked': get_requirements(locked=True),
    },
    setup_requires=['pytest-runner'],
    tests_require=[
        'pytest',
        'pytest-pep8',
        'pylint',
        'mock',
        'pytest-mock',
    ],
)



import logging
import entrypoints

from .base import BasePromoter


logger = logging.getLogger(__name__)


def load_promoters():
    """Return list of available promoters"""
    builtins = [
        RequestsPromoter,
        FirefoxPromoter,
        ScriptPromoter,
    ]
    return builtins + load_extensions()


def load_extensions():
    """Return list of fetcher promoters defined in Kibitzr extensions"""
    return [
        point.load()
        for point in entrypoints.get_group_all("kibitzr.fetcher")
    ]


class URLPromoter(BasePromoter):

    @staticmethod
    def is_applicable(conf):
        """Return whether this promoter is applicable for given conf"""
        return bool(conf.get('url'))

    def log_announcement(self):
        logger.info("Fetching %s at %s",
                    self.conf['name'], self.conf['url'])

    @staticmethod
    def needs_firefox(conf):
        return any(
            conf.get(key)
            for key in ('delay', 'scenario', 'form')
        )


class RequestsPromoter(URLPromoter):

    PRIORITY = 5  # default fallback

    def __init__(self, conf):
        super().__init__(conf)
        self._fetcher = None

    @classmethod
    def is_applicable(cls, conf):
        """Return whether this promoter is applicable for given conf"""
        return all((
            URLPromoter.is_applicable(conf),
            not cls.needs_firefox(conf),
        ))

    def fetch(self):
        from .simple import requests_fetcher  # pylint: disable=import-outside-toplevel
        super().fetch()
        if not self._fetcher:
            self._fetcher = requests_fetcher(self.conf)
        return self._fetcher()


class FirefoxPromoter(URLPromoter):

    PRIORITY = 15

    @classmethod
    def is_applicable(cls, conf):
        """Return whether this promoter is applicable for given conf"""
        return all((
            URLPromoter.is_applicable(conf),
            cls.needs_firefox(conf),
        ))

    def fetch(self):
        # pylint: disable=import-outside-toplevel
        from .browser.fetcher import firefox_fetcher
        super().fetch()
        return firefox_fetcher(self.conf)


class ScriptPromoter(BasePromoter):

    PRIORITY = 15

    @staticmethod
    def is_applicable(conf):
        """Return whether this promoter is applicable for given conf"""
        return all((
            'url' not in conf,
            'script' in conf,
        ))

    def log_announcement(self):
        logger.info("Fetching %r using script",
                    self.conf['name'])

    def fetch(self):
        from .script import fetch_by_script  # pylint: disable=import-outside-toplevel
        super().fetch()
        return fetch_by_script(self.conf)



from kibitzr.stash import Stash
from kibitzr.transformer.jinja_transform import JinjaTransform


class StashNotify(Stash):

    def __init__(self, conf, value):
        super().__init__()
        self.conf = conf
        self.data = value

    def render(self, report):
        context = None
        new_values = {}
        for key, code in self.data.items():
            transform = JinjaTransform(code, self.conf)
            if context is None:
                context = transform.context(report)
            ok, value = transform.render(report, context)
            if ok:
                new_values[key] = value
            else:
                return False, {}
        return True, new_values

    def save_report(self, report):
        ok, new_values = self.render(report)
        if ok:
            self.write(new_values)
    __call__ = save_report


notify_factory = StashNotify
